/*******************************************************************************
*
*    PID_SERVO_1_2MS
*
*    C/ microcontrolador  12F675
*
*    Autor: Cludio Lrios
*
*    Data: 06/12/2015
*
*    Usa sinal de 1-2 mseg proveniente de receptor comercial de controle remoto
*    e outro potencimetro para obter informao da posio atual (recpia)
*    Sendo as entradas analgicas:
*     AN3=sinal de 1-2 mseg (pino 4)
*     AN1=pot_recopia (pino 6)
*    Tendo como opo de driver:
*     tipo 0: ch0-Esquerda, ch1-Direita (ponte simples a 4 transistores)
*     tipo 1: ch0-on/off, ch1-direo do movimento (ponte integrada c/ pino
*             on/off e outro para Direco.
*
*    Uso didtico (circuito experimental)
*    Este arquivo  parte integrante do blog Larios.Tecnologia.ws
*
*******************************************************************************/
#include <12F675.h> //modelo do pic usado
#fuses WDT,INTRC_IO, NOCPD, NOPROTECT, NOMCLR, NOPUT, NOBROWNOUT//configurao
#use delay(clock=4000000) //frequencia de trabalho do pic
#use fast_io(a) //entrada e saidas feita pelo programador
#use rs232(baud=9600, xmit=PIN_A2)//pino para testes
//bytes
#byte ADRESH      = 0x1e
#byte ADRESL      = 0x9e
#byte ADCON0      = 0x1f
#byte ANSEL       = 0x9f
#byte TRISIO      = 0X85
#byte GPIO        = 0X05
#byte CMCON       = 0X19
#byte TMR1L       = 0X0E
#byte TMR1H       = 0X0F
#byte T1CON       = 0X10
#byte OPTION_REG  = 0X81

//bits
#bit    motorL   = 0x5.4
#bit    motorR   = 0x5.5
#bit    motorON  = 0x5.4
#bit    motorDIR = 0x5.5
#bit    in_sinal = 0x5.3   //pino 4  //sinal de comando 1 a 2 mseg (50hz)
#bit    TMR1ON   = 0X10.0  //on/off do timer1
#bit    GPIF     = 0X0B.0  //flag de mudana de estado
#bit    GIE      = 0X0B.7  //flag chave geral de interrupes
#bit    GPIE     = 0X0B.3  //flag chave interrupo de mudana
#bit    PEIE     = 0X0B.6  //flag chave interrupo de perifricos
#bit    TMR1IF   = 0X0C.0  //flag do overflow do tmr1
#bit    IOC3     = 0X96.3  //flag chave interrupo por mudana individual
#bit    GO_DONE  = 0X1F.1 //flag de inicio de converso do AD



#define  p_min   1000   //valor minimo do periodo recebido do tx (1 a 2 mseg)

//tipo de driver usado pelo hobista
#define tipo_de_driver 0 // ponte simples com 4 transistores (motorL e motorH)
//#define tipo_de_driver 1 // ponte integrada, com entrada 'on/off' e  'direo'

//Defines usados no PID
//ajuste valores abaixo para 'sintonizar' conforme o seu equipamento
#define Kp 0.4 * 16 //fator proporcional (erros geram valor do pwm proporcional as dimenses do erro)
                    // (Em excesso gera oscilaes no brao)
#define Ki 0.4 * 16 //fator integrativo (erros pequenos so corrigidos c/ o passar tempo)
                    //(excesso pode gerar ruidos no motor quando na posio desejada)
#define Kd 0.4 * 16 //fator derivativo (mudanas rpidas geram correes rpidas)
                    //(excesso pode contribuir para oscilaes no brao)

#define max_pwm 240 // maximo valor para pwm  (at 255)
#define max_integra 20000  //somatria maxima permitida
#define banda_morta 1 // limite de atuao
#define banda_morta_derivada 10 //limite para atuao da derivao
#define offset 30   //compensao no pwm

//macros atravs de 'define'
#define complementa(a) (a=((0xffff - a)+1))
#define Testa_sinal(i) (i&0x8000)

//variaveis globais
unsigned int1 f_sinal;
unsigned int16 r_sinal,pot_recopia,erro,pwm,aux,erro_anterior=0,deriva,integra=0;


/*******************************************************************************
           ROTINA DE INTERRUPO POR MUDANA DE ESTADO (PINO 4)
 ******************************************************************************/
 #INT_RA
void isr_le_pulso(){//le a largura do pulso (1 a 2 mseg)

     GPIO&=GPIO;
    if(!GPIF)return;//testa se foi realmente interrupo por mudana

    GPIF=0;//apaga flag
    if(in_sinal){//testa polaridade do evento de interrupo
        TMR1L=0;//zera
        TMR1H=0;//zera
        TMR1IF=0;//apaga flag do tmr1
        T1CON=1;//liga timer 1
        
    }
    else{
        if(TMR1IF){GPIO=0;   return; }//sai por erro
        T1CON=0;//desliga timer 1
        r_sinal=0;
        r_sinal=( (int16)TMR1H<<8)|TMR1L;//obtem o valor do pulso

        if(r_sinal>=p_min){
        r_sinal=r_sinal-p_min;// corrige posio
        r_sinal=(r_sinal>>1);//divide por 2
        }
        else{
        r_sinal=0;
        }
     }

}

/*******************************************************************************
                 MAIN - CONFIGURAES DE PORTAS E REGISTRADORES
*******************************************************************************/


void main(){
int a=0;

  TRISIO= 0b00001011;//define entradas e sadas
  ANSEL=0b00010011; //0:nc, 001:fosc/8, 0011:DDAA
  CMCON=7;//desliga comparadores

  #if tipo_de_driver==0
   motorL=0; motorR=0;
  #else
   motorON=0;
  #endif

   OPTION_REG=0B00001111;//WDT/128
   GPIE=1;//liga interrupo de mudana geral
   IOC3=1;//liga interrupo de mudana para o pino 4
   PEIE=1;//liga interrupo para os perifricos
   GPIF=0;//apaga flag
   GIE=1;//liga geral das interrupes
   delay_ms(100);//aguarda leitura do valor do pulso de entrada


/*******************************************************************************
                  LOOP PRINCIPAL DO PROGRAMA
*******************************************************************************/

for(;;){

  restart_wdt();//limpa wdt
  
//realiza converso analgica do potenciometro  recpia
//pot_recopia
  ADCON0=0b10000101; //1: Right, 0: vref=5v, 00:--, 01: canal=1, 1: go-done, 1: AD_ON
  delay_us(100);
  GO_DONE=1;//inicia
  while(GO_DONE);//aguarda converso terminar
  pot_recopia=(int16)ADRESH<<8|ADRESL;

/*******************************************************************************
          IMPLEMENTA CONTROLE  PID (PROPORCIONAL/INTEGRAL/DERIVATIVO)
*******************************************************************************/

//Kp
  pwm=0;
  erro=r_sinal-pot_recopia; //obtem o erro entre o 'pot objetivo' e o medido pelo 'pot_recopia'
  deriva=erro-erro_anterior; //deriva o erro da comparao
  erro_anterior=erro; //atualiza


  if((banda_morta+erro>4)||(banda_morta<erro)){
    if(Testa_sinal(erro)){ f_sinal=1; complementa(erro); }
    else{f_sinal=0;}
    
       
    aux=Kp * erro;
    if(f_sinal) complementa(aux);
    pwm+=aux;


//Kd
    if (Testa_sinal(deriva)) { f_sinal=1; complementa(deriva); }
    else{f_sinal=0;}
    if(deriva>banda_morta_derivada){//testa limite para ao
      aux=deriva * Kd; 
      if (f_sinal) complementa(aux);
      pwm+=aux;
    }
  }// if((banda_morta+erro>4)||(banda_morta<erro)){



//Ki
  if (Testa_sinal(integra)) { f_sinal=1; complementa(integra); }
  else{f_sinal=0;}
  if (integra>max_integra) integra=max_integra; //ajusta limite
  aux=integra;
  aux=aux>>4;//divide por 16
  aux=aux * Ki;//multiplica por fator
  aux=aux>>4;//divide por 16
  if (f_sinal) {complementa(aux); complementa(integra);}
  pwm+=aux;
  aux=pwm;

 
//ATIVA SADA DO MOTOR
  if(pwm){
     if(Testa_sinal(pwm)){// se negativo
       #if tipo_de_driver==0
        motorL=1;
       #else
        motorON=1;
        motorDIR=1;
       #endif
       complementa(aux);
  }
   else{//se positivo
    #if tipo_de_driver==0
     motorR=1;
    #else
     motorON=1;
     motorDIR=0;
    #endif
  }
  aux+=offset;
  if(aux>max_pwm) aux=max_pwm;
  }


//DURAO DO DUTY CICLE DO PWM (TEMPO ON)
  a=0;
  do{
    a++;
    delay_us(50);
    if ((int8)aux==a){
     #if tipo_de_driver==0
      motorL=0; motorR=0;
     #else
      motorON=0;
     #endif
    }
  }while(a<255);


 } //while(1)
}//main
